%{
//***************************************************************************
// Scanner flex specification
//***************************************************************************
/* Lex Definitions for a STAR File */

#ifdef YYLMAX
#undef YYLMAX
#endif
#define YYLMAX 65355
%}

a		     [aA]
b		     [bB]
d		     [dD]
e		     [eE]
g		     [gG]
l		     [lL]
o		     [oO]
p		     [pP]
s		     [sS]
t		     [tT]
v		     [vV]
eoln                 (\r\n|\r|\n)
Dollar	     	     \$
Data		     {d}{a}{t}{a}
Loop		     {l}{o}{o}{p}
Global	  	     {g}{l}{o}{b}{a}{l}
Stop		     {s}{t}{o}{p}
Save		     {s}{a}{v}{e}
Char		     [^\n\r]
Blank		     [\n\r \t]
NonBlank	     [!-~]
NonDQuote	     [^\n\r"]
NonSQuote	     [^\n\r']
NotUnderscore	     [^\n\r \t_]
Space	             [\t ]
Spaces               [\t ]+

%{
#include "portability.h"
#include <stdlib.h>
#include <string.h>
#include "scanner.h"
#include "ast.h"
#include "gram.tab.h" 	//token definitions etc.
#include "message.h"


int line_num=1;
int col_num=1;
bool err_is_printing = false;
int star_err_cnt = 0;
int last_loop_line = 0;
int last_loop_col = 0;
bool star_warning_check = true;
bool star_syntax_print = true;

int consecutive_datavalues = 0;

void fillLineAndColNum();
void handle_whitespace();
void handle_newline();
void count_newlines();
void debug_msg( const char* );
%}

%s INSEMIDELIM

%array

%%
<INITIAL>{Loop}_              {
		       fillLineAndColNum();
		       debug_msg( "lex: loopstart\n" );
                       return (LOOP);
		     }

<INITIAL>{Stop}_              {
		       fillLineAndColNum();
		       debug_msg( "lex: stop found\n" );
                       return (STOP);
		     }
                       
<INITIAL>{Data}_{NonBlank}+   {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: dataheading found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (DATA_HEADING);
		     }

<INITIAL>{Global}_            {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: global found\n" );
                       return (GLOBAL_HEADING);
		     }

<INITIAL>{Save}_{NonBlank}+   {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: saveheading found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (SAVE_HEADING);
		     }

<INITIAL>{Save}_              {
		       fillLineAndColNum();
		       debug_msg( "lex: saveend found\n" );
                       return (SAVE_END);
		     }
                       
<INITIAL>{Dollar}{NonBlank}+  {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: framecode found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (FRAME_CODE);
		     }

<INITIAL>_{NonBlank}+         {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: dataname found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (DATA_NAME);
		     }

<INITIAL>\"[^\n\r]*\"{Blank}    {
		       /* Got the longest "..." string in the line, but
		        * this string might contain smaller strings
			* in it, since regular expressions match the
			* longest thing they can.
			* Example:   _tag1 "val1" _tag2 "val2"
			* matched string is "val1" _tag2 "val2"
			* We want to break it apart if this is the
			* case, and push back the rest of the token
			* for further lexing.  (This ugliness is
			* required because STAR's spec says you can have
			* quotes in quotes provided they have no
			* whitespace after - that's a construct that
			* regular exp's cannot do, so we resort to this
			* icky read-too-far-and-pushback technique.)
			*/
		       char *ch;
		       int  prev_was_quote = 0;
		       for( ch = yytext+1 ; *ch != '\0' ; ch++ )
		       {
		           if( *ch == '\"' )
			   {   prev_was_quote = 1;
			   }
			   else
			   {   if( prev_was_quote )
			       {   prev_was_quote = 0;
				   if(    *ch == ' ' ||
				          *ch == '\t' ||
				          *ch == '\n' ||
					  *ch == '\r'  )
				   {   yyless( (ch - yytext) );
				       yytext[ ch - yytext ] = '\0';
				       ch--; /* pre-undo of for-loop's incr */
				   }
			       }
			   }
		       }

		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: doublequoted found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (DOUBLE_QUOTED_TEXT_STRING);
		     }

<INITIAL>\'[^\n\r]*\'{Blank}    {
		       /* Got the longest '...' string in the line, but
		        * this string might contain smaller strings
			* in it, since regular expressions match the
			* longest thing they can.
			* Example:   _tag1 'val1' _tag2 'val2'
			* matched string is 'val1' _tag2 'val2'
			* We want to break it apart if this is the
			* case, and push back the rest of the token
			* for further lexing.  (This ugliness is
			* required because STAR's spec says you can have
			* quotes in quotes provided they have no
			* whitespace after - that's a construct that
			* regular exp's cannot do, so we resort to this
			* icky read-too-far-and-pushback technique.)
			*/
		       char *ch;
		       int  prev_was_quote = 0;
		       for( ch = yytext+1 ; *ch != '\0' ; ch++ )
		       {
		           if( *ch == '\'' )
			   {   prev_was_quote = 1;
			   }
			   else
			   {   if( prev_was_quote )
			       {   prev_was_quote = 0;
				   if(    *ch == ' ' ||
				          *ch == '\t' ||
				          *ch == '\n' ||
					  *ch == '\r'    )
				   {   yyless( (ch - yytext) );
				       yytext[ ch - yytext ] = '\0';
				       ch--; /* pre-undo of for-loop's incr */
				   }
			       }
			   }
		       }

		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: singlequoted found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (SINGLE_QUOTED_TEXT_STRING);
		     }

<INITIAL>{eoln};                   { 
		       line_num++;
                       BEGIN INSEMIDELIM;  /* prevents "#" from being
		                            * interpeted as comments when
					    * in a semicolon string
					    */
		     }
<INSEMIDELIM>([^\n\r]*{eoln})({eoln}|([^;\n\r][^\n\r]*{eoln}))*;  {
                       count_newlines();
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(
		                          string(";") + string(yytext)  );

		       BEGIN INITIAL;  /* get out of INSEMIDELIM state. */

		       debug_msg( "lex: semiquoted found...(not printed)" );
		       debug_msg( yytext );
                       return (SEMICOLON_BOUNDED_TEXT_STRING);
		     }

<INITIAL>#[^\n\r]*/{eoln}   {  /**** THROW AWAY COMMENT READER ****/ }

<INITIAL>{NotUnderscore}{NonBlank}*   {
		       fillLineAndColNum();
		       yylval.TOKENval.strVal = new string(yytext);
		       debug_msg( "lex: nonquoted found: " );
		       debug_msg( yytext );
		       debug_msg( "\n" );
                       return (NON_QUOTED_TEXT_STRING);
		     }

<INITIAL>{Spaces}             {
                       handle_whitespace();
		     }

<INITIAL>{eoln}                   {
                       handle_newline();
		     }

<INITIAL>.                    {
                       scanner_error(line_num, col_num, "illegal character");
		       fillLineAndColNum();
		     }

%%
void fillLineAndColNum()
{
  yylval.TOKENval.lineNum = line_num;
  yylval.TOKENval.colNum = col_num;
  col_num += yyleng;
  return;
}

void handle_newline()
{
	col_num = yylval.TOKENval.colNum = 1;
	line_num++;
	yylval.TOKENval.lineNum = line_num;
	return;
}

void handle_whitespace()
{
	col_num += yyleng;
	yylval.TOKENval.colNum = col_num;
	yylval.TOKENval.lineNum = line_num;
	return;
}

void count_newlines()
{
  for (int i = 0; yytext[i] != 0; i++)
    if (yytext[i] == '\n')
      line_num++;

  col_num = 1;
}

void debug_msg( const char *print_this )
{
#ifdef DEBUG
    printf( "DEBUG:" );
    printf( "line:%d: %s, ", line_num, print_this );
    fflush( stdout );  /* debug msg MUST get out in case of a crash, */
                       /* else we can't trust it to tell us where it */
                       /* really crashed if it gets too buffered. */
#else
    char shut_up_lint = print_this[0];  // so it does something with the param.
#endif
}

void yyerror(char *msg)
{
    StarErrEntry  defaultError;

    defaultError.errID = STAR_UNKNOWN_ERR;
    defaultError.lineNum = line_num;
    defaultError.colNum = col_num;
    strncpy( defaultError.tokenStr, yytext, sizeof(defaultError.tokenStr)-1 );
    strncpy( defaultError.errStr, msg, sizeof(defaultError.errStr)-1 );
    char    tokenPrint[102];

    if( star_err_cnt == 0 )
    {
	// This is the first error ever encountered in
	// This particular parse.  If there was a previous
	// call to yyparse() in this living process, the
	// error list might still be populated from it,
	// so let's clear it before adding the errors for
	// this new parse.
	starErrReport.erase( starErrReport.begin(), starErrReport.end() );
    }
    else
    {
	if( (starErrReport.end()-1) -> errID == STAR_UNKNOWN_ERR )
	    if( star_syntax_print )
	        printf( "Parse error.\n" );
		                 // previous error was unknown and
		                 // its error message was never
				 // finished with finish_err_msg()
    }
    starErrReport.insert( starErrReport.end(), defaultError );
    star_err_cnt++;

    if(     index( yytext, '\n' ) != NULL ||
            strlen(yytext) > 100 )
	strcpy( tokenPrint, "[not printed]");
    else
	strcpy( tokenPrint, yytext );

    if(  strcmp( msg, "parse error" ) == 0 ||
         strcmp( msg, "Parse Error" ) == 0 ||
	 strcmp( msg, "Parse error" ) == 0 ||
	 strcmp( msg, "PARSE ERROR" ) == 0    )
    {
        err_is_printing = true;
	if( star_syntax_print )
	    printf( "ERR: at '%s' on line %d(%d): ",
		    tokenPrint, line_num, col_num );
    }
    else if(  strcmp( msg, "Warning" ) == 0 ||
	      strcmp( msg, "WARNING" ) == 0    )
    {
        err_is_printing = true;
	if( star_syntax_print )
	    printf( "WARN: at '%s' on line %d(%d): ",
		    tokenPrint, line_num, col_num );
    }
    else
    {
	if( star_syntax_print )
	    printf( "%s (line %d, col %d), reading '%s')\n",
		    msg, line_num, col_num, tokenPrint);
    }
}

void finish_err_msg( StarErrID id, const char *msg )
{

    // Update some of the fields on the most recent
    // error message (created by the last call to yyerror().)
    strcpy( (starErrReport.end() - 1)->errStr,  msg );
    (starErrReport.end() - 1)->errID = id;

    // Print the error msg:
    if( star_syntax_print )
    {
	printf( "%s\n", msg );
	fflush(stdout);
    }
    err_is_printing = false;
}

